@@ -114,6 +114,7 @@ SQLUSMALLINT MADB_supported_api[]=
 
 struct st_ma_connection_methods MADB_Dbc_Methods; /* declared at the end of file */
 
+
 my_bool CheckConnection(MADB_Dbc *Dbc)
 {
   if (!Dbc->mariadb)
@@ -338,8 +339,10 @@ SQLRETURN MADB_DbcGetAttr(MADB_Dbc *Dbc, SQLINTEGER Attribute, SQLPOINTER ValueP
                                           Dbc->CatalogName, strlen(Dbc->CatalogName), &Dbc->Error);
       ret= SQL_SUCCESS;
     }
-    if (StringLengthPtr)
+    if (StringLengthPtr != NULL)
+    {
       *StringLengthPtr= (SQLINTEGER)StrLen;
+    }
     return ret;
   }
   case SQL_ATTR_LOGIN_TIMEOUT:
@@ -437,6 +440,7 @@ MADB_Dbc *MADB_DbcInit(MADB_Env *Env)
   Connection->Methods= &MADB_Dbc_Methods;
   //CopyClientCharset(&SourceAnsiCs, &Connection->Charset);
   InitializeCriticalSection(&Connection->cs);
+  InitializeCriticalSection(&Connection->ListsCs);
   /* Not sure that critical section is really needed here - this init routine is called when
      no one has the handle yet */
   EnterCriticalSection(&Connection->Environment->cs);
@@ -506,13 +510,13 @@ SQLRETURN MADB_Dbc_GetCurrentDB(MADB_Dbc *Connection, SQLPOINTER CurrentDB, SQLI
   MADB_Stmt *Stmt;
   SQLRETURN ret;
   SQLLEN Size;
-  char Buffer[65 * sizeof(WCHAR)];
+  char Buffer[65 * sizeof(SQLWCHAR)];
 
   MADB_CLEAR_ERROR(&Connection->Error);
   ret= MA_SQLAllocHandle(SQL_HANDLE_STMT, (SQLHANDLE) Connection, (SQLHANDLE*)&Stmt);
   if (!SQL_SUCCEEDED(ret))
     return ret;
-  if (!SQL_SUCCEEDED(Stmt->Methods->ExecDirect(Stmt, (SQLCHAR *)"SELECT IF(DATABASE() IS NOT NULL,DATABASE(),'null')", SQL_NTS)) ||
+  if (!SQL_SUCCEEDED(Stmt->Methods->ExecDirect(Stmt, (char *)"SELECT IF(DATABASE() IS NOT NULL,DATABASE(),'null')", SQL_NTS)) ||
       !SQL_SUCCEEDED(Stmt->Methods->Fetch(Stmt)))
   {
     MADB_CopyError(&Connection->Error, &Stmt->Error);
@@ -533,12 +537,15 @@ end:
 }
 BOOL MADB_SqlMode(MADB_Dbc *Connection, enum enum_madb_sql_mode SqlMode)
 {
+  unsigned int ServerStatus;
+
+  mariadb_get_infov(Connection->mariadb, MARIADB_CONNECTION_SERVER_STATUS, (void*)&ServerStatus);
   switch (SqlMode)
   {
   case MADB_NO_BACKSLASH_ESCAPES:
-    return test(Connection->mariadb->server_status & SERVER_STATUS_NO_BACKSLASH_ESCAPES);
+    return test(ServerStatus & SERVER_STATUS_NO_BACKSLASH_ESCAPES);
   case MADB_ANSI_QUOTES:
-    return test(Connection->mariadb->server_status & SERVER_STATUS_ANSI_QUOTES);
+    return test(ServerStatus & SERVER_STATUS_ANSI_QUOTES);
   }
   return FALSE;
 }
@@ -600,6 +607,14 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
   {
     mysql_optionsv(Connection->mariadb, MYSQL_PLUGIN_DIR, Dsn->ConnCPluginsDir);
   }
+  else
+  {
+    const char *DefaultLocation= MADB_GetDefaultPluginsDir(Connection);
+    if (DefaultLocation != NULL)
+    {
+      mysql_optionsv(Connection->mariadb, MYSQL_PLUGIN_DIR, DefaultLocation);
+    }
+  }
 
   if (Dsn->ReadMycnf != '\0')
   {
@@ -622,14 +637,19 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
       cs_name= cs->csname;
     }
 
-    if (InitClientCharset(&Connection->Charset, MADB_IS_EMPTY(cs_name) ? "utf8" : cs_name))
+    if (InitClientCharset(&Connection->Charset, MADB_IS_EMPTY(cs_name) ? "utf8mb4" : cs_name))
     {
       /* Memory allocation error */
       MADB_SetError(&Connection->Error, MADB_ERR_HY001, NULL, 0);
       goto end;
     }
+    if (iOdbc() && strcmp(Connection->Charset.cs_info->csname, "swe7") == 0)
+    {
+      MADB_SetError(&Connection->Error, MADB_ERR_HY001, "Charset SWE7 is not supported with iODBC", 0);
+      goto end;
 
-    if (!Connection->IsAnsi)
+    }
+    if (!Connection->IsAnsi || iOdbc())
     {
       /* If application is not ansi, we should convert wchar into connection string */
       Connection->ConnOrSrcCharset= &Connection->Charset;
@@ -761,6 +781,20 @@ SQLRETURN MADB_DbcConnectDB(MADB_Dbc *Connection,
     mysql_optionsv(Connection->mariadb, MYSQL_OPT_SSL_CRLPATH, Dsn->SslCrlPath);
   }
 
+  if (!MADB_IS_EMPTY(Dsn->ServerKey))
+  {
+    mysql_optionsv(Connection->mariadb, MYSQL_SERVER_PUBLIC_KEY, Dsn->ServerKey);
+  }
+
+  if (!MADB_IS_EMPTY(Dsn->TlsPeerFp))
+  {
+    mysql_optionsv(Connection->mariadb, MARIADB_OPT_TLS_PEER_FP, (void*)Dsn->TlsPeerFp);
+  }
+  if (!MADB_IS_EMPTY(Dsn->TlsPeerFpList))
+  {
+    mysql_optionsv(Connection->mariadb, MARIADB_OPT_TLS_PEER_FP_LIST, (void*)Dsn->TlsPeerFpList);
+  }
+
   if (!mysql_real_connect(Connection->mariadb,
       Dsn->Socket ? "localhost" : Dsn->ServerName, Dsn->UserName, Dsn->Password,
         Dsn->Catalog && Dsn->Catalog[0] ? Dsn->Catalog : NULL, Dsn->Port, Dsn->Socket, client_flags))
@@ -1002,10 +1036,14 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
                       StringLengthPtr);
     break;
   case SQL_COLLATION_SEQ:
-    SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, 
-                                     (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar), 
-                                     Dbc->mariadb->charset->name, SQL_NTS, &Dbc->Error);
+  {
+    MY_CHARSET_INFO cs;
+    mariadb_get_infov(Dbc->mariadb, MARIADB_CONNECTION_MARIADB_CHARSET_INFO, (void*)&cs);
+    SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL,
+      (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar),
+      cs.name, SQL_NTS, &Dbc->Error);
     break;
+  }
   case SQL_COLUMN_ALIAS:
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, 
                            (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar), "Y", SQL_NTS, &Dbc->Error);
@@ -1298,7 +1336,7 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
     MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, SQL_GB_NO_RELATION, StringLengthPtr);
     break;
   case SQL_IDENTIFIER_CASE:
-    MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, SQL_IC_MIXED, StringLengthPtr);
+    MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, SQL_IC_MIXED, StringLengthPtr);
     break;
   case SQL_IDENTIFIER_QUOTE_CHAR:
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr, BUFFER_CHAR_LEN(BufferLength, isWChar), 
@@ -1367,7 +1405,7 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, 0, StringLengthPtr);
     break;
   case SQL_MAX_COLUMN_NAME_LEN:
-    MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, NAME_CHAR_LEN * SYSTEM_MB_MAX_CHAR_LENGTH, StringLengthPtr);
+    MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, NAME_CHAR_LEN * SYSTEM_MB_MAX_CHAR_LENGTH - 1, StringLengthPtr);
     break;
   case SQL_MAX_COLUMNS_IN_GROUP_BY:
     MADB_SET_NUM_VAL(SQLUSMALLINT, InfoValuePtr, 0, StringLengthPtr);
@@ -1415,7 +1453,7 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
   case SQL_MAX_STATEMENT_LEN:
     {
       size_t max_packet_size;
-      mariadb_get_infov(Dbc->mariadb, MARIADB_MAX_ALLOWED_PACKET, &max_packet_size);
+      mariadb_get_infov(Dbc->mariadb, MARIADB_MAX_ALLOWED_PACKET, (void*)&max_packet_size);
       MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, (SQLUINTEGER)max_packet_size, StringLengthPtr);
     }
     break;
@@ -1526,10 +1564,17 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
                                      "\\", SQL_NTS, &Dbc->Error);
     break;
   case SQL_SERVER_NAME:
-    SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr, 
-                                      BUFFER_CHAR_LEN(BufferLength, isWChar),
-                                     !(Dbc->mariadb) ? "" : Dbc->mariadb->host_info, SQL_NTS, &Dbc->Error);
+  {
+    const char *Host= "";
+    if (Dbc->mariadb)
+    {
+      mariadb_get_infov(Dbc->mariadb, MARIADB_CONNECTION_HOST, (void*)&Host);
+    }
+    SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr,
+      BUFFER_CHAR_LEN(BufferLength, isWChar),
+      Host, SQL_NTS, &Dbc->Error);
     break;
+  }
   case SQL_SPECIAL_CHARACTERS:
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr, 
                                       BUFFER_CHAR_LEN(BufferLength, isWChar),
@@ -1666,10 +1711,17 @@ SQLRETURN MADB_DbcGetInfo(MADB_Dbc *Dbc, SQLUSMALLINT InfoType, SQLPOINTER InfoV
     MADB_SET_NUM_VAL(SQLUINTEGER, InfoValuePtr, SQL_U_UNION | SQL_U_UNION_ALL, StringLengthPtr);
     break;
   case SQL_USER_NAME:
+  {
+    const char *User= "";
+    if (Dbc->mariadb)
+    {
+      mariadb_get_infov(Dbc->mariadb, MARIADB_CONNECTION_USER, (void *)&User);
+    }
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr,
                                      BUFFER_CHAR_LEN(BufferLength, isWChar), 
-                                     (Dbc->mariadb) ? Dbc->mariadb->user : "", SQL_NTS, &Dbc->Error);
+                                      User, SQL_NTS, &Dbc->Error);
     break;
+  }
   case SQL_XOPEN_CLI_YEAR:
     SLen= (SQLSMALLINT)MADB_SetString(isWChar ? &Dbc->Charset : NULL, (void *)InfoValuePtr,
                                      BUFFER_CHAR_LEN(BufferLength, isWChar),
@@ -1815,8 +1867,14 @@ SQLRETURN MADB_DriverConnect(MADB_Dbc *Dbc, SQLHWND WindowHandle, SQLCHAR *InCon
     goto error;
   }
  
-  if (!SQL_SUCCEEDED(DSNPrompt_Lookup(&DSNPrompt, Drv->SetupLibrary, Dbc)))
+  switch (DSNPrompt_Lookup(&DSNPrompt, Drv->SetupLibrary))
   {
+  case 0: break;
+  case MADB_PROMPT_NOT_SUPPORTED:
+    MADB_SetError(&Dbc->Error, MADB_ERR_HY000, "Prompting is not supported on this platform", 0);
+    goto error;
+  case MADB_PROMPT_COULDNT_LOAD:
+    MADB_SetError(&Dbc->Error, MADB_ERR_HY000, "Couldn't load the setup library", 0);
     goto error;
   }
 
@@ -1864,12 +1922,12 @@ end:
   {
     if (StringLength1 == SQL_NTS)
     {
-      StringLength1= (SQLSMALLINT)strlen(InConnectionString);
+      StringLength1= (SQLSMALLINT)strlen((const char*)InConnectionString);
     }
     if (OutConnectionString && BufferLength)
     {
       /* Otherwise we are supposed to simply copy incoming connection string */
-      strncpy_s((char *)OutConnectionString, BufferLength, InConnectionString, StringLength1);
+      strncpy_s((char *)OutConnectionString, BufferLength, (const char*)InConnectionString, StringLength1);
     }
     Length= StringLength1;
   }
